from math import log2
from typing import List, Hashable

from .empirical_probability import empirical_probability
from .abstract_measure import AbstractMeasure

class ShannonEntropyMeasure(AbstractMeasure):
    def generate(self, values: List[Hashable]) -> float:
        probabilities = empirical_probability(values=values)
        entropy = -sum(probability * log2(probability) for probability in probabilities.values() if probability > 0)
        return abs(entropy)


import unittest
import os
from dotenv import load_dotenv

load_dotenv()
openai_api_key = os.getenv("OPENAI_API_KEY")

from langdiversity.models import OpenAIModel

class TestOpenAI(unittest.TestCase):
    def test_basic(self):
        openai_model = OpenAIModel(openai_api_key=openai_api_key, model="gpt-3.5-turbo")
        response = openai_model.generate("hi", count=2)
        print(response)


import unittest

from langdiversity.measures import ShannonEntropyMeasure

class TestShannonEntropyMeasure(unittest.TestCase):
    def test_basic(self):
        shannon_entropy_measure = ShannonEntropyMeasure()
        output = shannon_entropy_measure.generate(values=[5, 5, 5, 4.5, 3])
        self.assertAlmostEqual(output, 1.370950594454669)


import re

class NumberExtractor:

    @staticmethod
    def extract_decimals_from_text(text: str) -> str:
        preprocessed_text = NumberExtractor.__preprocess(text)
        extracted_strings = re.findall(r'-?\d+\.?\d*\/-?\d+\.?\d*|-?\d+\.?\d*', preprocessed_text)
        postprocessed_text = NumberExtractor.__postprocess(extracted_strings)
        return postprocessed_text
    
    def __preprocess(text: str) -> str: 
        text = NumberExtractor.__remove_commas_between_numbers(text)
        return text
    
    def __postprocess(text: str) -> str:
        text = [NumberExtractor.__remove_leading_zeros(extracted_string) for extracted_string in text]
        return text

    def __remove_leading_zeros(text: str) -> str:
        while len(text) > 1 and text.startswith('0') and text[1].isdigit(): text = text[1:]
        return text

    def __remove_commas_between_numbers(text: str) -> str: 
        """
        Removes commas that are directly in between numbers.\n
        For example,\n
        0, 0 --> 0, 0 will stay the same.\n
        0,0 --> 00 will have its comma removed.\n
        """
        
        return re.sub(r'(?<=\d)\,(?=\d)', '', text)



from typing import List, Hashable

from .empirical_probability import empirical_probability
from .abstract_measure import AbstractMeasure

class GiniImpurityMeasure(AbstractMeasure):
    def generate(self, values: List[Hashable]) -> float:
        probabilities = empirical_probability(values=values)
        gini_impurity = 1 - sum(probability ** 2 for probability in probabilities.values())
        return gini_impurity


import unittest

from langdiversity.measures import empirical_probability

class TestEmpiricalProbability(unittest.TestCase):
    def test_basic(self):
        output = empirical_probability([1, 1, 1, 2, 2])
        self.assertEqual(output, {1: 0.6, 2: 0.4})

    def test_single_distinct_value(self):
        output = empirical_probability([1, 1, 1, 1, 1])
        self.assertEqual(output, {1: 1})

    def test_empty_array(self):
        output = empirical_probability([])
        self.assertEqual(output, {})


import re

from .number_extractor import NumberExtractor

# =======================================
# This class is concerned with evaluating
# ChatGPT responses. It contains functions used to
# evaluate very specific datasets
# =======================================
def last_letters(response: str, answer: str):
    response = extract_last_letters(response)
    return response == answer

def extract_last_letters(response: str):
    response = response.lower()
    response = extract_answer(response)
    response = re.findall(r'[a-zA-Z]+', response)

    if len(response) == 0: return ''
    return response[0]


def extract_math_answer(response: str):
    response = response.lower()
    response = extract_answer(response)
    response = NumberExtractor.extract_decimals_from_text(response)

    responses = []
    for r in response:
        try: responses.append(eval(r))
        except: pass
    
    return frozenset(responses)


def extract_multi_choice_answer(response: str):
    response = response.lower()
    response = extract_answer(response)
    response = re.findall(r'a|b|c|d|e', response, flags=re.IGNORECASE)

    if len(response) == 0: return ''
    return response[0]

def extract_answer(response: str) -> str:
    response = __clean_response(response)

    extracted_answer = re.findall(r'the answer is.*', response, flags=re.IGNORECASE)
    if len(extracted_answer) == 0: return ''

    extracted_answer = extracted_answer[0]
    extracted_answer = re.sub(r'the answer is', '', extracted_answer, count=1, flags=re.IGNORECASE)

    if len(re.findall(r'the answer is', extracted_answer, flags=re.IGNORECASE)) > 0: return extract_answer(extracted_answer)
    return extracted_answer.strip()

def __clean_response(response: str) -> str:
    response = re.sub(r' +', ' ', response)
    return response



from typing import List, Dict, Union

import json

class PromptSelection:
    def __init__(self, data: List[Dict[str, Union[str, List[str], float]]], selection: str = "min"):
        valid_selections = ["min", "max"]
        if selection not in valid_selections:
            raise ValueError(
                "Invalid selection type. Expected one of %s" % valid_selections
            )
        self.data = data
        self.selection = selection

    def select(self):
        if not self.data:
            raise ValueError("No data to select from.")

        # Determine the target diversity value based on the selection type
        if self.selection == "min":
            target_diversity = min(item['diversity'] for item in self.data)
        else:
            target_diversity = max(item['diversity'] for item in self.data)

        # Filter the data list to return all items with the target diversity value
        selected_items = [item for item in self.data if item['diversity'] == target_diversity]

        # Prepare a dictionary to hold the selected prompts and diversity value
        results = {
            "selected_prompts": [item['prompt'] for item in selected_items],
            "diversity": target_diversity
        }

        return results 




from typing import List

from langdiversity.extras.spinner import loading_spinner

from langdiversity.measures import AbstractMeasure
from langdiversity.models import AbstractBaseModel

class DiversityMeasureCollector:
    def __init__(self, model: AbstractBaseModel, diversity_measure: AbstractMeasure, num_responses: int = 1):
        self.model = model
        self.diversity_measure = diversity_measure
        self.num_responses = num_responses
        self.data = []  # A list to store the data (prompt, responses, diversity measure)

    def collect(self, prompts: List[str], verbose: bool = False):
        total_prompts = len(prompts)
        for i, prompt in enumerate(prompts):
            with loading_spinner(f"Collecting {self.num_responses} responses...", current_step=i + 1, total_steps=total_prompts):
                responses = self.model.generate(prompt, self.num_responses)
            with loading_spinner("Performing diversity measure calculations...", current_step=i + 1, total_steps=total_prompts):
                diversity = self.diversity_measure.generate(responses)
            if verbose:
                print(f"Prompt {i + 1}: {prompt}")
                print(f"Responses: {', '.join(responses)}")  
                print(f"Diversity Measure ({self.diversity_measure.__class__.__name__}): {diversity}")
            self.data.append(
                {"prompt": prompt, "responses": responses, "diversity": diversity}
            )



import itertools
import sys
import threading
import time
from contextlib import contextmanager

@contextmanager
def loading_spinner(label, current_step=None, total_steps=None):
    spinner_done = False

    def spinner_function():
        spinner = itertools.cycle(["-", "/", "|", "\\"])
        while not spinner_done:
            prefix = f"[{current_step}/{total_steps}] " if current_step is not None and total_steps is not None else ""
            sys.stdout.write(f'\r{prefix}{label} {next(spinner)}')
            sys.stdout.flush()
            time.sleep(0.1)
        print("\r", end="")
        print(f"{prefix}{label} ✓")

    spinner_thread = threading.Thread(target=spinner_function)
    spinner_thread.start()

    try:
        yield
    finally:
        spinner_done = True
        spinner_thread.join()





import abc
from typing import List, Optional, Callable

from langchain.schema import HumanMessage, SystemMessage
from langchain.chat_models import ChatOpenAI

from .abstract_base_model import AbstractBaseModel


class OpenAIModel(AbstractBaseModel):
    def __init__(
        self,
        openai_api_key: str,
        model: str = "gpt-3.5-turbo",
        extractor: Optional[Callable] = None,
        *args,
        **kwargs
    ):
        self.openai_api_key = openai_api_key
        self.kwargs = kwargs
        if "n" in self.kwargs:
            del self.kwargs["n"]

        self.model = model
        self.extractor = extractor

    def generate(self, message: str, count: int = 1):
        messages = [[SystemMessage(content=""), HumanMessage(content=message)]]

        chat_openai = ChatOpenAI(
            **self.kwargs, n=count, model=self.model, openai_api_key=self.openai_api_key
        )

        response = chat_openai.generate(
            messages=messages,
        )
        parsed_response = self.__parse_response(response)

        # Apply the extractor if provided
        if self.extractor:
            parsed_response = [self.extractor(resp) for resp in parsed_response]

        return parsed_response

    def __parse_response(self, response: dict):
        choices = response.generations[0]
        messages = [choice.text for choice in choices]
        return messages



import unittest
import os
from dotenv import load_dotenv
from langdiversity.models import OpenAIModel
from langdiversity.utils import DiversityCalculator, PromptSelection
from langdiversity.parser import extract_last_letters

load_dotenv()
openai_api_key = os.getenv("OPENAI_API_KEY")

class TestPromptSelection(unittest.TestCase):

    def setUp(self):
        self.openai_model = OpenAIModel(openai_api_key=openai_api_key, model="gpt-3.5-turbo", extractor=extract_last_letters)
        self.diversity_calculator = DiversityCalculator()
        self.prompt_selector = PromptSelection()

    def test_prompt_selection(self):
        ll_question = "Take the last letter of each word in \"Tal Evan Lesley Sidney\" and concatenate them."
        prompt = f"At the end, say 'the answer is [put the concatenated word here]'.\nQuestion: {ll_question}.\n "
        
        all_responses = []
        all_scores = []
        for _ in range(10):  # Repeating the process 10 times
            responses = self.openai_model.generate(prompt, count=5)
            diversity_scores = self.diversity_calculator.calculate(responses, measures=['entropy', 'gini'])
            all_responses.append(responses)
            all_scores.append(diversity_scores)
        
        selected_prompts_max = self.prompt_selector.find_extreme_scores(all_scores, measures=['entropy', 'gini'], selection_method='max')
        selected_prompts_min = self.prompt_selector.find_extreme_scores(all_scores, measures=['entropy', 'gini'], selection_method='min')
        
        # Shows all responses and scores
        for i, (responses, scores) in enumerate(zip(all_responses, all_scores)):
            print(f"Run {i+1}:")
            print("Responses:", responses)
            print("Scores:", scores)
            print("----------")
        
        print("Selected Prompts (Max):", selected_prompts_max)
        print("Selected Prompts (Min):", selected_prompts_min)



from langdiversity.extras.spinner import loading_spinner

class DiversityCalculator:
    def __init__(self, measures=[]):
        self.measures = measures

    def calculate(self, values):
        results = {}
        for measure in self.measures:
            display_name = measure.__class__.__name__
            with loading_spinner(f"Calculating '{display_name}' for {len(values)} values..."):
                results[display_name] = measure.generate(values)
                
        return results



from setuptools import find_packages, setup

# Read the contents of README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='langdiversity',
    packages=find_packages(exclude=['tests']),
    version='1.1.3',
    description='A tool to elevate your language models with insightful diversity metrics.',
    long_description=long_description,  
    long_description_content_type="text/markdown",  
    author='Noel Ngu, Nathaniel Lee',                        
    author_email='nngu2@asu.edu, nlee51@asu.edu',             
    url='https://github.com/lab-v2/langdiversity',  
    license='BSD 3-clause',
    install_requires=[
        'aiohttp==3.8.5',
        'aiosignal==1.3.1',
        'annotated-types==0.5.0',
        'async-timeout==4.0.3',
        'attrs==23.1.0',
        'certifi==2023.7.22',
        'charset-normalizer==3.2.0',
        'dataclasses-json==0.5.14',
        'frozenlist==1.4.0',
        'greenlet==2.0.2',
        'idna==3.4',
        'langchain==0.0.281',
        'langsmith==0.0.33',
        'marshmallow==3.20.1',
        'multidict==6.0.4',
        'mypy-extensions==1.0.0',
        'numexpr==2.8.5',
        'numpy==1.25.2',
        'openai==0.28.0',
        'packaging==23.1',
        'pydantic==2.3.0',
        'pydantic_core==2.6.3',
        'PyYAML==6.0.1',
        'requests==2.31.0',
        'SQLAlchemy==2.0.20',
        'tenacity==8.2.3',
        'typing-inspect==0.9.0',
        'typing_extensions==4.7.1',
        'urllib3==2.0.4',
        'yarl==1.9.2',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',

    ],
    python_requires='>=3.9',
)



# import unittest
# import os
# from dotenv import load_dotenv
# from langdiversity.models import OpenAIModel
# from langdiversity.utils import DiversityCalculator
# from langdiversity.parser import extract_last_letters

# load_dotenv()
# openai_api_key = os.getenv("OPENAI_API_KEY")

# #TODO: The extractor usage might be sus???
# class TestDiversityMeasuresOnLLMResponses(unittest.TestCase):

#     def setUp(self):
#         self.openai_model = OpenAIModel(openai_api_key=openai_api_key, model="gpt-3.5-turbo", extractor=extract_last_letters)
#         self.diversity_calculator = DiversityCalculator()

#     def test_langdiversity(self):
#         ll_question = "Take the last letter of each word in \"Tal Evan Lesley Sidney\" and concatenate them."
#         prompt = f"At the end, say 'the answer is [put the concatenated word here]'.\nQuestion: {ll_question}.\n "
#         responses = self.openai_model.generate(prompt, count=5)
        
#         diversity_scores = self.diversity_calculator.calculate(responses, measures=['entropy', 'gini'])
        
#         print("LLM Responses:", responses)
#         print("Diversity Scores:", diversity_scores)



from collections import Counter
from typing import List, Hashable

def empirical_probability(values: List[Hashable]) -> List[float]:
    """
    Calculates the probability that a discrete random variable will take 
    on a particular value for each unique value in the list.
    """
    value_counts = Counter(values)
    total_elements = len(values)
    probabilities = {value: count / total_elements for value, count in value_counts.items()}
    return probabilities